class IncompatibleDimensionsException(Exception):
    def __init__(self):
        self.message = "Previous layer shape is incompatible with this layer's shape!"
        super().__init__(self.message)




"""
KMer library that handles KMer counting and KMer sequence encoding.
"""
import re
import multiprocessing as mp
import numpy as np
from tqdm import tqdm


class KMerCounter:
    """
    KMer Counter class that can convert DNA/RNA sequences into sequences of kmers or kmer
    frequency tables. Wraps logic for Python multiprocessing using jobs and chunksize.
    """
    def __init__(self, k: int, jobs=1, chunksize=1, debug=False, quiet=False):
        """
        KMer Counter class that can convert DNA/RNA sequences into sequences of kmers or kmer
        frequency tables. Wraps logic for Python multiprocessing using jobs and chunksize.
        @param debug: disables multiprocessing to allow better tracebacks.
        """
        self.k = k
        self.jobs = jobs
        self.chunksize = chunksize
        self.debug = debug
        self.quiet = quiet
        alphabet = np.array(['A', 'C', 'G', 'T', 'U'])
        self.alphabet_pattern = re.compile(f'[^{"".join(alphabet)}]')

        # Make a lookup table with an entry for every possible data byte
        self.lookup_table = np.zeros(256, dtype=np.uint32)
        # A = 0 is implied by np.zeros
        self.lookup_table[ord('C')] = 1
        self.lookup_table[ord('G')] = 2
        self.lookup_table[ord('T')] = 3
        self.lookup_table[ord('U')] = 3  # U = T

    def _split_str(self, seq: str) -> list[str]:
        """
        Splits the input using the self.alphabet_pattern regex pattern.
        Filters out anything smaller than k. Acts as a generator.
        """
        for p in self.alphabet_pattern.split(seq):
            if len(p) >= self.k:
                yield np.array([p]).view(np.uint32)

    def _seq_to_kmers(self, seq: np.ndarray) -> np.ndarray:
        """
        Convert a sequence of base pair bytes to a sequence of integer kmers.
        SEQ MUST NOT HAVE BASE PAIRS OUTSIDE OF ACGT/U!
        """
        binary_converted = self.lookup_table[seq]  # Convert seq to integers
        stride = np.lib.stride_tricks.sliding_window_view(binary_converted, self.k)
        # If every base pair is a unique two-bit code, a kmer is the concatenation of these codes
        # In other words, kmer = sum(val << (kmer_len - idx) * 2 for idx, val in kmer_window)
        # where val is a 2-bit sequence
        kmers = np.copy(stride[:, -1])  # Start with a new array to store the sum
        for i in range(stride.shape[1] - 2, -1, -1):  # Iterate over columns in reverse
        # Add this column << (kmer_len - idx) * 2 to sum
            kmers += stride[:, i] << (stride.shape[1] - i - 1) * 2
        return kmers

    def str_to_kmers(self, seq: str) -> np.ndarray:
        """
        Convert a string sequence to integer kmers. Works around invalid base pairs.
        Remember that values of 0 and 1 are also used by keras for pad and OOV tokens! Be sure to
        add 2 before passing in a kmer sequence directly to a model.
        """
        # Split given string into valid parts and concatenate resulting kmer sequences for each part
        return np.concatenate([self._seq_to_kmers(i) for i in self._split_str(seq)])

    def _seq_to_kmer_counts(self, seq: np.ndarray) -> np.ndarray:
        """
        Convert an array of base pair bytes to an array of kmer frequencies.
        SEQ MUST NOT HAVE BASE PAIRS OUTSIDE OF ACGT/U!
        """
        kmers = self._seq_to_kmers(seq)
        kmer_counts = np.zeros(4 ** self.k)
        uniques, counts = np.unique(kmers, return_counts=True)
        kmer_counts[uniques] = counts
        return kmer_counts

    def str_to_kmer_counts(self, seq: str) -> np.ndarray:
        """
        Convert a string sequence to kmer counts. Works around invalid base pairs.
        """
        # Split given string into parts and take sum of each kmer's total occurrences in each part
        return np.sum([self._seq_to_kmer_counts(i) for i in self._split_str(seq)], axis=0)

    def _apply(self, func: callable, seqs: np.ndarray, jobs=None, chunksize=None,
               silence=False) -> list:
        """
        Avoids duplication of logic for kmer sequence/count generation.
        """
        jobs = jobs or self.jobs
        chunksize = chunksize or self.chunksize
        if not self.debug and jobs > 1:
            with mp.Pool(jobs) as p:
                it = p.imap(func, seqs, chunksize=chunksize)
                return list(it if self.quiet or silence else tqdm(it, total=len(seqs)))
        else:
            return [func(i) for i in (seqs if self.quiet or silence else tqdm(seqs))]

    def kmer_sequences(self, seqs: list[str], **kwargs) -> list[list[int]]:
        """
        Generate kmer sequences for a given array of string sequences.
        Sequences do not need to be uniform lengths. Invalid/unknown base pairs will be ignored.
        Remember that values of 0 and 1 are also used by keras for pad and OOV tokens! Be sure to
        add 2 before passing in a kmer sequence directly to a model.
        """
        return self._apply(self.str_to_kmers, seqs, **kwargs)

    def kmer_counts(self, seqs: list[str], **kwargs) -> np.ndarray:
        """
        Generate kmer counts for a given array of string sequences.
        Sequences do not need to be uniform lengths. Invalid/unknown base pairs will be ignored.
        """
        return np.array(self._apply(self.str_to_kmer_counts, seqs, **kwargs))


class Nucleotide_AA(KMerCounter):
    """
    Convert a nucleotide sequence to a "comprehensive" amino acid sequence containing all three
    possible start shifts. 'J', 'O', 'U', 'X' are letters unused in standard amino acid one-letter
    codes. We thus use X to represent all three kinds of stop codon.
    """
    AA_LOOKUP = np.array([
        'K',  # AAA
        'N',  # AAC
        'K',  # AAG
        'N',  # AAU
        'T', 'T', 'T', 'T',  # AC*
        'R',  # AGA
        'S',  # AGC
        'R',  # AGG
        'S',  # AGU
        'I',  # AUA
        'I',  # AUC
        'M',  # AUG
        'I',  # AUU
        'Q',  # CAA
        'H',  # CAC
        'Q',  # CAG
        'H',  # CAU
        'P', 'P', 'P', 'P',  # CC*
        'R', 'R', 'R', 'R',  # CG*
        'L', 'L', 'L', 'L',  # CU*
        'E',  # GAA
        'D',  # GAC
        'E',  # GAG
        'D',  # GAU
        'A', 'A', 'A', 'A',  # GC*
        'G', 'G', 'G', 'G',  # GG*
        'V', 'V', 'V', 'V',  # GU*
        'X',  # UAA (STOP)
        'Y',  # UAC
        'X',  # UAG (STOP)
        'Y',  # UAU
        'S', 'S', 'S', 'S',  # UC*
        'X',  # UGA  (STOP)
        'C',  # UGC
        'W',  # UGG
        'C',  # UGU
        'L',  # UUA
        'F',  # UUC
        'L',  # UUG
        'F'  # UUU
    ])

    def __init__(self, **kwargs):
        super().__init__(3, **kwargs)

    def _kmer_seq_to_aa(self, seq: np.ndarray) -> str:
        return ''.join(self.AA_LOOKUP[seq])

    def transform(self, seqs: list[str]) -> np.ndarray:
        """
        Convert the given nucleotide sequences to comprehensive amino acid sequences.
        @param seqs: nucleotide sequences to convert.
        """
        kmers = self.kmer_sequences(seqs)
        if not self.quiet:
            print('Converting to amino acid sequences...')
        return np.array(self._apply(self._kmer_seq_to_aa, kmers), dtype=str)




"""
Library for input compression before data is passed into a model.
"""
import os
import shutil
import pickle
import copy
import multiprocessing as mp
import numpy as np
from tqdm import tqdm
import tensorflow as tf
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA as SKPCA, IncrementalPCA
from .kmers import KMerCounter


class KMerCountCompressor(KMerCounter):
    #pylint: disable=unused-argument
    """
    Abstract Compressor class used for compressing input data.
    """
    _SAVE_EXCLUDE_VARS = []
    def __init__(self, counter: KMerCounter, compress_to: int):
        super().__init__(counter.k, jobs=counter.jobs, chunksize=counter.chunksize,
                         debug=counter.debug, quiet=counter.quiet)
        self.compress_to = compress_to
        self.fit_called = False

    def save(self, savedir: str):
        """
        Save the Compressor to the filesystem.
        """
        shutil.rmtree(savedir, ignore_errors=True)
        os.makedirs(savedir)

        to_pkl = copy.copy(self)  # Efficient shallow copy for pickling
        for i in self._SAVE_EXCLUDE_VARS:  # Don't pickle attrs in _SAVE_EXCLUDE_VARS
            delattr(to_pkl, i)

        with open(os.path.join(savedir, 'compressor.pkl'), 'wb') as f:
            pickle.dump(to_pkl, f)

    @staticmethod
    def load(savedir: str):
        """
        Load the Compressor from the filesystem.
        """
        if not os.path.exists(savedir) or not os.path.exists(savedir):
            raise ValueError("Directory doesn't exist!")
        if 'compressor.pkl' not in os.listdir(savedir):
            raise ValueError('compressor.pkl is necessary!')
        with open(os.path.join(savedir, 'compressor.pkl'), 'rb') as f:
            obj = pickle.load(f)
        # pylint: disable=protected-access
        obj._load_special(savedir)
        return obj

    def _load_special(self, savedir: str):
        """
        Load any special variables from the savedir for this object. Called by Compressor.load().
        """

    def fit(self, data: np.ndarray):
        """
        Fit the compressor to the given data.
        @param data: data to fit to.
        @param quiet: whether to print output
        """
        self.fit_called = True

    def raw_transform(self, data: np.ndarray) -> np.ndarray:
        """
        The most basic transform operation, after kmer counting. Must be implemented.
        """
        return data

    def _transform_with_kmer_counts(self, data: np.ndarray) -> np.ndarray:
        data = self.kmer_counts(data, silence=True, jobs=1, chunksize=1)
        return self.raw_transform(data)

    def transform(self, data: np.ndarray, silence=False) -> np.ndarray:
        # pylint: disable=unused-argument
        """
        Compress an array of data elements.
        @param data: data to compress.
        @param silence: additional option to silence output of this function.
        @return np.ndarray: compressed data.
        """
        return data

    def inverse_transform(self, data: np.ndarray, silence=False) -> np.ndarray:
        # pylint: disable=unused-argument
        """
        Decodes the compressed data back to original.
        @param data: data to decode.
        @param silence: additional option to silence output of this function.
        @return np.ndarray: uncompressed data.
        """
        return data


class _PCACompressor(KMerCountCompressor):
    """
    Abstract PCA compressor, conserves code.
    """
    def __init__(self, pca, *args, jobs=1, chunksize=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.pca = pca
        self.scaler = StandardScaler()
        self.compress_jobs = jobs
        self.compress_chunksize = chunksize

    def _batch_data(self, data: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        fully_batchable_data = data[:len(data) - len(data) % self.compress_chunksize]
        full_batches = np.reshape(fully_batchable_data,
                                  (-1, self.compress_chunksize, *fully_batchable_data.shape[1:]))
        last_batch = data[len(data) - len(data) % self.compress_chunksize:]
        return full_batches, last_batch

    def _mp_map_over_batches(self, fn: callable, data: np.ndarray, silence=False) -> np.ndarray:
        full_batches, last_batch = self._batch_data(data)
        with mp.Pool(self.compress_jobs) as p:
            it = p.imap_unordered(fn, full_batches) if self.quiet or silence else tqdm(
                p.imap_unordered(fn, full_batches), total=len(full_batches))
            result = list(it)
        if len(last_batch) > 0:
            result.append(fn(last_batch))
        return np.concatenate(result) if len(result) > 0 and isinstance(result[0], np.ndarray) \
                else result

    def raw_transform(self, data: np.ndarray, silence=False) -> np.ndarray:
        return self.pca.transform(self.scaler.transform(data))

    def transform(self, data: np.ndarray, silence=False) -> np.ndarray:
        if isinstance(data[0], str):  # Calling on list of strings
            data = np.array(data, dtype=object)
            transform_fn = self._transform_with_kmer_counts
        else:  # Calling on kmer counts
            transform_fn = self.raw_transform
        return self._mp_map_over_batches(transform_fn, data, silence)

    def _raw_inverse_transform(self, data: np.ndarray) -> np.ndarray:
        return self.scaler.inverse_transform(self.pca.inverse_transform(data))

    def inverse_transform(self, data: np.ndarray, silence=False):
        return self._mp_map_over_batches(self._raw_inverse_transform, data, silence)


class KMerCountPCA(_PCACompressor):
    """
    Use PCA to compress input data. Suppoorts parallelization on transform, not fit.
    """
    def __init__(self, counter: KMerCounter, n_components: int, **kwargs):
        pca = SKPCA(n_components=n_components)
        super().__init__(pca, counter, n_components, **kwargs)

    def fit(self, data: np.ndarray):
        super().fit(data)
        self.pca.fit(self.scaler.fit_transform(data))


class KMerCountIPCA(_PCACompressor):
    """
    Use PCA to compress the input data. Supports fit parallelization over multiple CPUs.
    """
    def __init__(self, counter: KMerCounter, n_components: int, **kwargs):
        """
        Uses jobs, chunksize defined by KMerCounter
        """
        super().__init__(None, counter, n_components, **kwargs)
        self.pca = IncrementalPCA(n_components=n_components, batch_size=self.compress_chunksize)

    def fit(self, data: np.ndarray):
        super().fit(data)
        if not self.quiet:
            print(f'Fitting IPCA Compressor using CPUs: {self.compress_jobs}...')
        data = self.scaler.fit_transform(data)
        full_batches, last_batch = self._batch_data(data)
        if len(last_batch) < self.compress_to:  # Drop last batch if not enough data
            last_batch = full_batches[-1]
            full_batches = full_batches[:-1]
        self._mp_map_over_batches(self.pca.partial_fit, np.concatenate(full_batches))
        # Use normal fit on last batch so sklearn doesn't trigger a fit not called error
        self.pca.fit(last_batch)


class KMerCountAE(KMerCountCompressor):
    """
    Train an autoencoder to compress the input data.
    """
    _SAVE_EXCLUDE_VARS = ['encoder', 'decoder', 'ae']
    def __init__(self, counter: KMerCounter, inputs: tf.keras.layers.Layer, reprs: tf.keras.layers.Layer,
                 outputs: tf.keras.layers.Layer, repr_size: int, batch_size: int, loss='mse',
                 epoch_limit=100, patience=2, val_split=.1, **kwargs):
        """
        Create an encoder/decoder autoencoder pair.
        Encoder: inputs=inputs, outputs=reprs
        Decoder: inputs=reprs, outputs=outputs
        Autoencoder: inputs=inputs, outputs=outputs
        epoch_limit, patience: modify fit() behavior, can be overridden in fit()
        """
        super().__init__(counter, repr_size, **kwargs)
        self.batch_size = batch_size
        self.encoder = tf.keras.Model(inputs=inputs, outputs=reprs)
        self.decoder = tf.keras.Model(inputs=reprs, outputs=outputs)
        self.ae = tf.keras.Model(inputs=inputs, outputs=outputs)
        self.ae.compile(optimizer='adam', loss=loss)
        self.epoch_limit = epoch_limit
        self.patience = patience
        self.val_split = val_split

    @classmethod
    def auto(cls, counter: KMerCounter, data: np.ndarray, repr_size: int, output_activation=None,
            **kwargs):
        """
        Automatically generate an autoencoder based on the input data. Recommended way to create
        an AECompressor.
        """
        inputs = tf.keras.layers.Input(data.shape[1:])
        x = tf.keras.layers.Dense(data.shape[-1], activation='relu')(inputs)
        reprs = tf.keras.layers.Dense(repr_size)(x)
        x = tf.keras.layers.Dense(data.shape[-1], activation='relu')(reprs)
        outputs = tf.keras.layers.Dense(data.shape[-1], activation=output_activation)(x)
        return cls(counter, inputs, reprs, outputs, repr_size, **kwargs)

    def save(self, savedir: str):
        super().save(savedir)
        self.encoder.save(os.path.join(savedir, 'encoder.h5'))
        self.decoder.save(os.path.join(savedir, 'decoder.h5'))
        self.ae.save(os.path.join(savedir, 'ae.h5'))

    def _load_special(self, savedir: str):
        self.encoder = tf.keras.models.load_model(os.path.join(savedir, 'encoder.h5'))
        self.decoder = tf.keras.models.load_model(os.path.join(savedir, 'decoder.h5'))
        self.ae = tf.keras.models.load_model(os.path.join(savedir, 'ae.h5'))

    def summary(self):
        """
        Print a summary of this autoencoder.
        """
        self.ae.summary()

    def fit(self, data: np.ndarray, epoch_limit=None, patience=None, val_split=None):
        """
        Train the autoencoder model on the given data. Uses early stopping to end training.
        """
        super().fit(data)
        epoch_limit = epoch_limit or self.epoch_limit
        patience = patience or self.patience
        val_split = val_split or self.val_split
        if not self.quiet:
            print('Training AE Compressor...')
        else:
            tf.keras.utils.disable_interactive_logging()
        self.ae.fit(data, data, epochs=epoch_limit, batch_size=self.batch_size, callbacks=[
            tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=patience),
        ], validation_split=val_split)
        if self.quiet:
            tf.keras.utils.enable_interactive_logging()

    def transform(self, data: np.ndarray, silence=False) -> np.ndarray:
        return self.encoder(data) if self.quiet or silence else \
            self.encoder.predict(data, batch_size=self.batch_size)

    def inverse_transform(self, data: np.ndarray, silence=False) -> np.ndarray:
        return self.decoder(data) if self.quiet or silence else \
            self.decoder.predict(data, batch_size=self.batch_size)




"""
Automated pipelines for sequence representation generation.
"""
import os
import shutil
import pickle
import multiprocessing as mp

from tqdm import tqdm
import numpy as np
import pandas as pd
import tensorflow as tf
import matplotlib.pyplot as plt
import scipy.stats as st
from scipy.spatial.distance import euclidean
from sklearn.neighbors import BallTree

from .dataset_builder import DatasetBuilder, SILVA_header_parser, COVID_header_parser
from .visualize import repr_scatterplot
from .kmers import KMerCounter, Nucleotide_AA
from .kmer_compression import KMerCountPCA, KMerCountIPCA, KMerCountAE, KMerCountCompressor
from .encoders import ModelBuilder
from .comparative_encoder import ComparativeEncoder, Decoder, DenseDecoder, LinearDecoder
from .distance import IncrementalDistance, EditDistance, Distance, Cosine, Euclidean


class Pipeline:
    """
    An abstract automated pipeline for sequence representation generation.
    """
    AVAILABLE_DECODERS = {
        'default': Decoder,
        'dense': DenseDecoder,
        'linear': LinearDecoder
    }
    def __init__(self, model=None, decoder=None, dataset=None, preproc_reprs=None, reprs=None,
                 quiet=False, random_seed=None):
        self.model, self.decoder = model, decoder
        self.dataset, self.preproc_reprs, self.reprs = dataset, preproc_reprs, reprs
        self.quiet = quiet
        self.index = None
        self.rng = np.random.default_rng(seed=random_seed)
        tf.random.set_seed(random_seed)  # Set global random seed

    def set_decoder(self, decoder, *args, **kwargs):
        """
        Manually set the decoder.
        """
        self.decoder = decoder if isinstance(decoder, Decoder) else \
            self.AVAILABLE_DECODERS[decoder](*args, **kwargs)

    def load_dataset(self, paths: list[str], header_parser='None', trim_to=0, max_rows=None):
        """
        Load a dataset into memory from a list of FASTA files.
        """
        if not isinstance(header_parser, str):
            builder = DatasetBuilder(header_parser)
        elif header_parser == 'SILVA':
            builder = DatasetBuilder(SILVA_header_parser)
        elif header_parser == 'COVID':
            builder = DatasetBuilder(COVID_header_parser)
        else:
            builder = DatasetBuilder()
        self.dataset = builder.from_fasta(paths, max_rows=max_rows)
        self.dataset.replace_unknown_nucls()
        if trim_to:
            self.dataset.trim_seqs(trim_to)

    def sample_data(self, size: int):
        """
        Randomly samples the Dataset and all representations down to size. Irreversible.
        """
        sample = self.rng.permutation(size)[:size]
        self.dataset = self.dataset.iloc[sample]
        if self.preproc_reprs is not None:
            self.preproc_reprs = self.preproc_reprs[sample]
        if self.reprs is not None:
            self.reprs = self.reprs[sample]

    # Subclass must override.
    @staticmethod
    def preprocess_seqs(seqs) -> np.ndarray:
        """
        Preprocesses a list of sequences.
        @param seqs: Sequences to preprocess.
        @return np.ndarray: Returns an array of preprocessed sequences.
        """
        if isinstance(seqs, list):
            return np.array(seqs)
        if isinstance(seqs, pd.Series):
            return seqs.to_numpy()
        return seqs

    # Must be implemented by subclass, super method must be called by implementation.
    # This super method preprocesses the dataset into self.preproc_reprs.
    # This variable is used to determine whether fit was called and to avoid preprocessing the
    # dataset twice between fit and transform_dataset. Returns indices of unique sequences.
    def fit(self, **kwargs):
        """
        Fit the model to the dataset.
        """
        if self.dataset is None:
            raise ValueError('Must load dataset before calling fit!')
        if not self.quiet:
            print('Preprocessing dataset...')
        _, unique_inds = np.unique(self.dataset['seqs'], return_index=True)
        self.preprocess_dataset(**kwargs)
        return unique_inds

    def fit_decoder(self, distance_on: np.ndarray, transform_batch_size: int, **kwargs):
        """
        Fit the decoder based on the model's representations.
        """
        if self.dataset is None:
            raise ValueError('Must load dataset before fitting decoder!')
        # Always transform dataset in case encoder has changed.
        if not self.quiet:
            print('Transforming dataset...')
        self.transform_dataset(transform_batch_size)
        if not self.quiet:
            print("Training Distance Decoder...")
        self.decoder.fit(self.reprs, distance_on, **kwargs)

    def _fit_decoder(self, dec_args: dict, batch_size: int, *args, **kwargs):
        """
        Adds defaults to dec_args before calling fit_decoder.
        """
        dec_args = dec_args or {}
        if 'transform_batch_size' not in dec_args:
            # Use the encoder's batch_size to transform
            dec_args['transform_batch_size'] = batch_size
        self.fit_decoder(*args, **(kwargs | dec_args))

    def _fit_called_check(self):
        if self.preproc_reprs is None:
            raise ValueError('Fit must be called before transform!')

    def plot_training_history(self, savepath=None):
        """
        Plot the training history of the trained model. Converts 1 - r loss into r^2.
        """
        data = self.model.history['loss']
        plt.plot(np.arange(len(data)), data)
        plt.title('ComparativeEncoder Training Loss History')
        plt.xlabel('Epoch')
        plt.ylabel('Model Loss')
        if savepath:
            plt.savefig(savepath)
        plt.show()

    def transform(self, seqs: list, batch_size: int) -> list:
        """
        Transform an array of string sequences to learned representations.
        @param seqs: List of string sequences to transform.
        @return list: Sequence representations.
        """
        self._fit_called_check()
        return self.model.transform(self.preprocess_seqs(seqs), batch_size)

    def preprocess_dataset(self, **kwargs) -> np.ndarray:
        """
        Preprocesses all sequences in dataset.
        """
        self.preproc_reprs = self.preprocess_seqs(self.dataset['seqs'], **kwargs)

    def transform_dataset(self, batch_size: int, **kwargs) -> np.ndarray:
        """
        Transforms the loaded dataset into representations. Saves as self.reprs and returns result.
        Deletes any existing search tree.
        """
        self._fit_called_check()
        self.reprs = self.model.transform(self.preproc_reprs, batch_size, **kwargs)
        self.index = None  # Delete existing search tree because we assume reprs have changed.
        return self.reprs

    def save(self, savedir: str):
        """
        Save the Pipeline to the given directory.
        """
        shutil.rmtree(savedir, ignore_errors=True)
        os.makedirs(savedir)
        models_path = os.path.join(savedir, 'models')
        if self.model is not None:
            self.model.save(os.path.join(models_path, 'encoder'))
        # pylint: disable=unidiomatic-typecheck
        if self.decoder is not None and not type(self.decoder) == Decoder:
            self.decoder.save(os.path.join(models_path, 'decoder'))
        if self.preproc_reprs is not None:
            np.save(os.path.join(savedir, 'preproc_reprs.npy'), self.preproc_reprs)
        if self.reprs is not None:
            np.save(os.path.join(savedir, 'reprs.npy'), self.reprs)

    @classmethod
    def load(cls, savedir: str, strategy=None, quiet=False, **custom_kwargs):
        """
        Load a Pipeline from the savedir. Keyword arguments passed to ComparativeEncoder load.
        """
        # pylint: disable=broad-exception-caught
        if not os.path.exists(savedir):
            raise ValueError("Directory doesn't exist!")
        contents = os.listdir(savedir)
        kwargs = cls._load_special(savedir)
        kwargs['quiet'] = quiet
        if 'models' in contents:
            thisdir = os.path.join(savedir, 'models')
            if os.path.exists(os.path.join(thisdir, 'encoder')):
                kwargs['model'] = ComparativeEncoder.load(os.path.join(thisdir, 'encoder'),
                                                          strategy=strategy, quiet=quiet)
            else:
                print('Warning: encoder missing!')
            if os.path.exists(os.path.join(thisdir, 'decoder')):
                kwargs['decoder'] = Decoder.load(os.path.join(thisdir, 'decoder'), quiet=quiet)
            else:
                kwargs['decoder'] = Decoder()
        else:
            print('Warning: models missing!')
        if 'preproc_reprs.npy' in contents:
            try:
                kwargs['preproc_reprs'] = np.load(os.path.join(savedir, 'preproc_reprs.npy'),
                                                  allow_pickle=True)
            except Exception as e:
                print('Warning: exception in loading preproc_reprs: ' + e)
        if 'reprs.npy' in contents:
            try:
                kwargs['reprs'] = np.load(os.path.join(savedir, 'reprs.npy'), allow_pickle=True)
            except Exception as e:
                print('Warning: exception in loading preproc_reprs: ' + e)
        kwargs.update(custom_kwargs)
        return cls(**kwargs)

    @staticmethod
    def _load_special(savedir: str) -> dict:
        # pylint: disable=unused-argument
        """
        Returns a dictionary of all loaded special constructor arguments for this Pipeline.
        """
        return {}

    def _reprs_check(self):
        """
        Wraps logic to check that reprs exist.
        """
        if self.reprs is None:
            raise ValueError('transform_dataset must be called first!')

    def visualize_axes(self, x: int, y: int, **kwargs):
        """
        Visualizes two axes of the dataset representations on a simple scatterplot.
        @param x: which axis to use as x.
        @param y: which axis to use as y.
        @param kwargs: Accepts additional keyword arguments for visualize.repr_scatterplot().
        """
        self._reprs_check()
        repr_scatterplot(np.stack([self.reprs[:, x], self.reprs[:, y]], axis=1), **kwargs)

    def visualize_2D(self, **kwargs):
        """
        Visualizes 2D dataset as a scatterplot. Keyword arguments to repr_scatterplot are accepted.
        """
        self._reprs_check()
        if len(self.reprs.shape) != 2 or self.reprs.shape[1] != 2:
            raise ValueError('Incompatible representation dimensions!')
        self.visualize_axes(0, 1, **kwargs)

    def search(self, query: list[str], n_neighbors=1,
               **kwargs) -> tuple[np.ndarray, list[pd.Series]]:
        """
        Search the dataset for the most similar sequences to the query. Accepts keyword arguments to
        Decoder.transform().
        @param query: List of string sequences to find similar sequences to.
        @param n_neighbors: Number of neighbors to find for each sequence. Defaults to 1.
        @return np.ndarray: Search results.
        """
        self._reprs_check()
        if self.model.embed_dist == 'euclidean':
            if self.index is None:  # If index hasn't been created, create it.
                if not self.quiet:
                    print('Creating search index...')
                self.index = BallTree(self.reprs)
            query_enc = self.transform([query], 1)
            dists, ind = self.index.query(query_enc, k=n_neighbors)
            matches = self.dataset.iloc[ind[0]]
            return self.decoder.transform(dists[0], **kwargs).flatten(), matches
        if self.model.embed_dist == 'hyperbolic':  # TODO: BALL TREE
            query_enc = self.transform([query], 1)
            x = np.repeat(query_enc, len(self.reprs), axis=0)
            dists = self.decoder.embed_dist_calc.transform_multi(x, self.reprs)
            s = np.argsort(dists)[:n_neighbors]
            return self.decoder.transform(dists, **kwargs)[s], self.dataset.iloc[s]
        raise ValueError('Invalid embedding distance!')  # Should never happen

    def evaluate(self, **kwargs):
        """
        Evaluate the performance of the model by seeing how well we can predict true sequence
        dissimilarity from encoding distances.
        @param sample_size: Number of sequences to use for evaluation. All in dataset by default.
        @return np.ndarray, np.ndarray: predicted distances, true distances
        """
        self._reprs_check()
        return self.decoder.evaluate(self.reprs, self.preproc_reprs, **kwargs)


class KMerCountsPipeline(Pipeline):
    """
    Automated pipeline using KMer Counts. Optionally compresses input data before training model.
    """
    DISTS = {
        'cosine': Cosine,
        'euclidean': Euclidean,
        'edit': EditDistance
    }
    def __init__(self, counter=None, compressor=None, **kwargs):
        super().__init__(**kwargs)
        self.counter = counter
        self.K_ = self.counter.k if self.counter else None
        self.repr_size_ = self.model.properties['repr_size'] if self.model else None
        self.compressor = compressor

    def create_kmer_counter(self, K: int, **kwargs):
        """
        Add a KMerCounter to this KMerCountsPipeline.
        """
        self.counter = KMerCounter(K, quiet=self.quiet, **kwargs)
        self.K_ = self.counter.k

    def create_compressor(self, compressor: str, repr_size=0, fit_sample_frac=1, **init_args):
        """
        Add a Compressor to this KMerCountsPipeline.
        """
        if not self.counter:
            raise ValueError('KMerCounter needs to be created before running! \
                             Use create_kmer_counter().')
        if compressor == 'None' or not compressor:
            self.compressor = KMerCountCompressor(self.counter, 4 ** self.K_, jobs=self.counter.jobs,
                                                  chunksize=self.counter.chunksize)
            return
        elif compressor not in ['PCA', 'IPCA', 'AE']:
            raise ValueError('Invalid Compressor Provided')
        sample = self.rng.permutation(len(self.dataset))[:int(len(self.dataset) * fit_sample_frac)]
        print('Counting k-mers in compressor fit sample...')
        sample = self.counter.kmer_counts(self.dataset['seqs'].to_numpy()[sample])
        compress_to = repr_size or 4 ** self.K_ // 10 * 2
        if compressor == 'PCA':
            self.compressor = KMerCountPCA(self.counter, compress_to, **init_args)
        elif compressor == 'IPCA':
            self.compressor = KMerCountIPCA(self.counter, compress_to, **init_args)
        elif compressor == 'AE':
            if 'batch_size' not in init_args:
                raise ValueError('batch_size must be provided in init_args for AE compressor!')
            self.compressor = KMerCountAE.auto(self.counter, sample, compress_to, **init_args)
            if not self.quiet:
                print('AE Compressor Summary:')
                self.compressor.summary()
        # pylint: disable=unidiomatic-typecheck
        # Strict type check needed here for this conditional
        if self.model is not None and not self.quiet and type(self.compressor) != Compressor:
            print('Creating a compressor after the model is not recommended! Consider running  \
                  create_model again.')
        print('Fitting compressor...')
        self.compressor.fit(sample)

    def create_model(self, repr_size=2, embed_dist='euclidean', depth=3, decoder='linear', dist='cosine',
                     distribute_strategy=None, dist_args=None, dec_args=None, embed_dist_args=None):
        """
        Create a Model for this KMerCountsPipeline. Uses all available GPUs.
        """
        # Argument validation
        if not self.counter:
            raise ValueError('KMerCounter needs to be created before running! \
                             Use create_kmer_counter().')
        if dist not in self.DISTS:
            raise ValueError('Invalid argument: dist. Must be one of "cosine", "edit", "euclidean"')
        dist = self.DISTS[dist](quiet=self.quiet, **(dist_args or {}))

        if not self.compressor:  # Create default (blank) compressor if needed
            self.create_compressor('None')

        dec_args = dec_args or {}
        if decoder == 'dense' and 'batch_size' not in dec_args:
            raise ValueError('Must pass "batch_size" in param dec_args for dense decoder')
        if 'embed_dist_args' not in dec_args:
            dec_args['embed_dist_args'] = embed_dist_args
        self.set_decoder(decoder, dist=dist, embed_dist=embed_dist, **dec_args)  # Create decoder
        builder = ModelBuilder((self.compressor.compress_to,),
                               distribute_strategy=distribute_strategy or
                               tf.distribute.MirroredStrategy())
        builder.dense(self.compressor.compress_to, depth=depth)
        self.model = ComparativeEncoder.from_model_builder(builder, dist=dist, repr_size=repr_size,
                                                           quiet=self.quiet, embed_dist=embed_dist,
                                                           random_seed=self.rng.integers(2**32))
        if not self.quiet:
            self.model.summary()

    def preprocess_seqs(self, seqs: list[str], **kwargs) -> np.ndarray:
        return self.compressor.transform(seqs, **kwargs)

    def fit(self, batch_size: int, preproc_args=None, dec_args=None, **kwargs):
        """
        Fit model to loaded dataset. Accepts keyword arguments for ComparativeEncoder.fit().
        Automatically calls create_model() with default arguments if not already called.
        """
        if not self.model:
            self.create_model()

        # Always preprocess (with compression) since this is necessary for model.
        unique_inds = super().fit(**(preproc_args or {}))
        # If k-mer count based distance and not compressing
        if isinstance(self.model.distance, (Euclidean, Cosine)):
            if type(self.compressor) == KMerCountCompressor:
                distance_on = self.preproc_reprs  # Feed kmer counts as input
            else:  # kmer based distance with compression
                # Use an IncrementalDistance to reduce memory usage
                self.model.distance = IncrementalDistance(self.model.distance, self.counter)
                self.decoder.distance = IncrementalDistance(self.decoder.distance, self.counter)
                distance_on = self.dataset['seqs'].to_numpy()
        else:
            # IncrementalDistance takes sequences as input
            distance_on = self.dataset['seqs'].to_numpy()

        self.model.fit(batch_size, self.preproc_reprs[unique_inds],
                       distance_on=distance_on[unique_inds], **kwargs)
        self._fit_decoder(dec_args, batch_size, distance_on)


    def save(self, savedir: str):
        super().save(savedir)
        with open(os.path.join(savedir, 'counter.pkl'), 'wb') as f:
            pickle.dump(self.counter, f)
        if self.compressor is not None:
            self.compressor.save(os.path.join(savedir, 'compressor'))

    @classmethod
    def load(cls, *args, **kwargs):
        return super().load(*args, strategy=tf.distribute.MirroredStrategy(), **kwargs)

    @staticmethod
    def _load_special(savedir: str):
        result = {}
        contents = os.listdir(savedir)
        if 'counter.pkl' not in contents:
            raise ValueError('counter is necessary!')
        with open(os.path.join(savedir, 'counter.pkl'), 'rb') as f:
            result['counter'] = pickle.load(f)
        if 'compressor' in contents:
            result['compressor'] = KMerCountCompressor.load(os.path.join(savedir, 'compressor'))
        return result

class SequencePipeline(Pipeline):
    """
    Abstract sequence alignment estimator.
    """
    VOCAB = []  # MUST be defined by subclass
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def create_model(self, res='low', seq_len=.9, repr_size=2, embed_dist='euclidean', decoder='linear',
                     dec_args=None, dist_args=None, embed_dist_args=None, **kwargs):
        """
        Create a model for the Pipeline.
        @param res: Resolution of the model's encoding output. Available options are:
            'low' (default): Basic dense neural network operating on top of learned embeddings for
            input sequences.
            'medium': Convolutional layer operating on 1/4 the length of input sequences.
            'high': Convolutional layer + attention block operating on 1/4 the length of input
            sequences.
            'ultra': Convolutional layer + attention block operating on full length of input
            sequences.
        @param seq_len: Specifies input length of sequences to model. Three possibilities:
            seq_len == None: Auto-detect the maximum sequence length and use as model input size.
            0 < seq_len < 1: Ensure that this fraction of the total dataset is NOT truncated.
            seq_len >= 1: Trim and pad directly to this length.
        @param repr_size: Number of dimensions in output encodings (default 2).
        @param embed_dist: Embedding space geometry, 'euclidean' or 'hyperbolic'
        @param dist_args: Arguments for distance metric (jobs, chunksize)
        @param **kwargs: Everything else passed to ComparativeEncoder.from_model_builder
        """
        if (seq_len is None or seq_len < 1) and self.dataset is None:
            raise ValueError('Dataset must be loaded before autodetection of sequence length!')
        if seq_len is not None and seq_len < 1:
            target_zscore = st.norm.ppf(seq_len)
            lengths = self.dataset['seqs'].apply(len)
            mean = np.mean(lengths)
            std = np.std(lengths)
            seq_len = int(target_zscore * std + mean)

        dist_args = dist_args or {}
        dist = EditDistance(quiet=self.quiet, **dist_args)

        dec_args = dec_args or {}
        if decoder == 'dense' and 'batch_size' not in dec_args:
            raise ValueError('Must pass "batch_size" in param dec_args for dense decoder')
        if 'embed_dist_args' not in dec_args:
            dec_args['embed_dist_args'] = embed_dist_args
        self.set_decoder(decoder, dist=dist, embed_dist=embed_dist, **dec_args)  # Create decoder

        if res == 'low':
            self.model = self.low_res_model(seq_len, repr_size=repr_size, dist=dist,
                                            random_seed=self.rng.integers(2**32),
                                            embed_dist=embed_dist, **kwargs)
        elif res == 'medium':
            self.model = self.medium_res_model(seq_len, repr_size=repr_size, dist=dist,
                                            random_seed=self.rng.integers(2**32),
                                            embed_dist=embed_dist, **kwargs)
        elif res == 'high':
            self.model = self.high_res_model(seq_len, repr_size=repr_size, dist=dist,
                                            random_seed=self.rng.integers(2**32),
                                            embed_dist=embed_dist, **kwargs)
        elif res == 'ultra':
            self.model = self.ultra_res_model(seq_len, repr_size=repr_size, dist=dist,
                                            random_seed=self.rng.integers(2**32),
                                            embed_dist=embed_dist, **kwargs)
        else:
            raise ValueError('Invalid argument: res must be one of "low", "medium", "high", "ultra"')

        if not self.quiet:
            self.model.summary()

    @classmethod
    def low_res_model(cls, seq_len: int, compress_factor=1, depth=3, **kwargs):
        """
        Basic dense neural network operating on top of learned embeddings for input sequences.
        """
        builder = ModelBuilder.text_input(cls.VOCAB, embed_dim=8, max_len=seq_len,
                                          distribute_strategy=tf.distribute.MirroredStrategy())
        builder.transpose()
        builder.dense(seq_len // compress_factor, depth=depth)
        builder.transpose()
        model = ComparativeEncoder.from_model_builder(builder, **kwargs)
        return model

    @classmethod
    def medium_res_model(cls, seq_len: int, compress_factor=4, conv_filters=16, conv_kernel_size=6,
                         dense_depth=3, **kwargs):
        """
        Convolutional layer operating on 1/4 the length of input sequences.
        """
        builder = ModelBuilder.text_input(cls.VOCAB, embed_dim=12, max_len=seq_len,
                                          distribute_strategy=tf.distribute.MirroredStrategy())
        builder.transpose()
        if dense_depth:
            builder.dense(seq_len, depth=dense_depth)
        builder.dense(seq_len // compress_factor)
        builder.transpose()
        builder.conv1D(conv_filters, conv_kernel_size, seq_len // compress_factor)
        model = ComparativeEncoder.from_model_builder(builder, **kwargs)
        return model

    @classmethod
    def high_res_model(cls, seq_len: int, compress_factor=4, conv_filters=32, conv_kernel_size=8,
                       attn_heads=2, dense_depth=3, **kwargs):
        """
        Convolutional layer + attention block operating on 1/4 the length of input sequences.
        """
        builder = ModelBuilder.text_input(cls.VOCAB, embed_dim=16, max_len=seq_len,
                                          distribute_strategy=tf.distribute.MirroredStrategy())
        builder.transpose()
        if dense_depth:
            builder.dense(seq_len, depth=dense_depth)
        builder.dense(seq_len // compress_factor)
        builder.transpose()
        builder.conv1D(conv_filters, conv_kernel_size, seq_len // compress_factor * 4)
        builder.reshape((*builder.shape()[:-2], builder.shape()[-1] // 4, 4))
        builder.attention(attn_heads, seq_len // compress_factor)
        model = ComparativeEncoder.from_model_builder(builder, **kwargs)
        return model

    @classmethod
    def ultra_res_model(cls, seq_len: int, compress_factor=1, conv_filters=64, conv_kernel_size=16,
                        attn_heads=4, dense_depth=3, **kwargs):
        """
        Convolutional layer + attention block operating on full length of input sequences.
        """
        builder = ModelBuilder.text_input(cls.VOCAB, embed_dim=20, max_len=seq_len,
                                          distribute_strategy=tf.distribute.MirroredStrategy())
        builder.transpose()
        if dense_depth:
            builder.dense(seq_len, depth=dense_depth)
        builder.dense(seq_len // compress_factor)
        builder.transpose()
        builder.conv1D(conv_filters, conv_kernel_size, seq_len // compress_factor * 4)
        builder.reshape((*builder.shape()[:-2], builder.shape()[-1] // 4, 4))
        builder.attention(attn_heads, seq_len // compress_factor)
        model = ComparativeEncoder.from_model_builder(builder, **kwargs)
        return model

    def fit(self, batch_size=256, dec_args=None, **kwargs):
        """
        Fit model to loaded dataset. Accepts keyword arguments for ComparativeEncoder.fit().
        Automatically calls create_model() with default arguments if not already called.
        """
        if not self.model:
            print('Warning: using default low-res model...')
            self.create_model()
        unique_inds = super().fit()

        self.model.fit(batch_size, self.preproc_reprs[unique_inds], **kwargs)
        self._fit_decoder(dec_args, batch_size, self.preproc_reprs)

    @classmethod
    def load(cls, *args, **kwargs):
        return super().load(*args, strategy=tf.distribute.MirroredStrategy(), **kwargs)


class DNASequencePipeline(SequencePipeline):
    """
    Edit distance estimator for DNA sequences.
    """
    VOCAB = ['A', 'C', 'G', 'T']


class RNASequencePipeline(SequencePipeline):
    """
    Edit distance estimator for RNA sequences.
    """
    VOCAB = ['A', 'C', 'G', 'U']


class HomologousSequencePipeline(SequencePipeline):
    """
    Edit distance of 3 possible forward reading frames.
    """
    VOCAB = np.unique(Nucleotide_AA.AA_LOOKUP)
    def __init__(self, converter=None, **kwargs):
        super().__init__(**kwargs)
        self.converter = converter

    def create_converter(self, *args, **kwargs):
        """
        Create a Nucleotide_AA converter for the Pipeline. Directly wraps constructor.
        """
        self.converter = Nucleotide_AA(*args, **kwargs)

    def preprocess_seqs(self, seqs: list[str]):
        if self.converter is None:
            print('Warning: default converter being used...')
            self.create_converter()
        return self.converter.transform(seqs)

    def save(self, savedir: str):
        super().save(savedir)
        with open(os.path.join(savedir, 'converter.pkl'), 'wb') as f:
            pickle.dump(self.converter, f)

    @staticmethod
    def _load_special(savedir: str):
        result = {}
        contents = os.listdir(savedir)
        if 'converter.pkl' in contents:
            with open(os.path.join(savedir, 'converter.pkl'), 'rb') as f:
                result['converter'] = pickle.load(f)
        return result




"""
ComparativeEncoder module, trains a model comparatively using distances.
"""
import os
import shutil
import pickle
import json
import time
import math

import numpy as np
import pandas as pd
from scipy.spatial.distance import euclidean
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error, r2_score, mean_absolute_percentage_error
import tensorflow as tf
import tensorflow.math as M

from .encoders import ModelBuilder
from .distance import Hyperbolic, Euclidean, Cosine


def _run_tf_fn(init_message=None, print_time=False):
    def fit_dec(fn):
        def fit_output_mgmt(self, *args, **kwargs):
            start_time = time.time()
            if self.quiet:
                tf.keras.utils.disable_interactive_logging()
            elif init_message:
                print(init_message)
            result = fn(self, *args, **kwargs)
            if self.quiet:
                tf.keras.utils.enable_interactive_logging()
            elif print_time:
                print(f'Total time taken: {time.time() - start_time} seconds.')
            return result
        return fit_output_mgmt
    return fit_dec


def _prepare_tf_dataset(x, y, batch_size):
    dataset = tf.data.Dataset.from_tensor_slices((x, y) if y is not None else x)
    dataset = dataset.batch(batch_size)

    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = \
        tf.data.experimental.AutoShardPolicy.DATA
    dataset = dataset.with_options(options)
    return dataset


class _NormalizedDistanceLayer(tf.keras.layers.Layer):
    """
    Adds a scaling parameter that's set to 1 / average distance on the first iteration.
    """
    def __init__(self, init_scaling=True, **kwargs):
        super().__init__(**kwargs)
        self.scaling_param = self.add_weight(
            shape=(),
            initializer=tf.keras.initializers.Ones(),
            trainable=True,
            name="scaling_param"
        )
        self.scaling_initialized = tf.Variable(not init_scaling, trainable=False)

    def norm(self, dists):
        """
        Normalize the distances with scaling and set scaling if first time.
        """
        if not self.scaling_initialized:
            self.scaling_param.assign(1 / tf.reduce_mean(dists))
            self.scaling_initialized.assign(True)
        return dists * (self.scaling_param + tf.keras.backend.epsilon())


class EuclideanDistanceLayer(_NormalizedDistanceLayer):
    """
    This layer computes the distance between its two prior layers.
    """
    def call(self, a, b):
        return self.norm(tf.reduce_sum(tf.square(a - b), -1))


class HyperbolicDistanceLayer(_NormalizedDistanceLayer):
    def call(self, a, b):
        """
        Computes hyperbolic distance in Poincaré ball model.
        """
        # Partially adapted from https://github.com/kousun12/tf_hyperbolic
        sq_norm = lambda v: tf.clip_by_value(
            tf.reduce_sum(v ** 2, axis=-1),
            clip_value_min=tf.keras.backend.epsilon(),
            clip_value_max=1 - tf.keras.backend.epsilon()
        )
        numerator = tf.reduce_sum(tf.pow(a - b, 2), axis=-1)
        denominator_a = 1 - sq_norm(a)
        denominator_b = 1 - sq_norm(b)
        frac = 1 + 2 * numerator / (denominator_a * denominator_b)
        hyperbolic_distance = tf.math.acosh(frac) + tf.keras.backend.epsilon()
        return self.norm(hyperbolic_distance)


class ComparativeModel:
    """
    Abstract ComparativeModel class. Stores some useful common functions.
    """
    def __init__(self, v_scope='model', dist=None, embed_dist='euclidean', model=None,
                 strategy=None, history=None, quiet=False, properties=None, random_seed=None,
                 **kwargs):
        self.strategy = strategy or tf.distribute.get_strategy()
        self.distance = dist
        self.quiet = quiet
        self.properties = {} if properties is None else properties
        self.history = history or {}
        self.embed_dist = embed_dist
        self.rng = np.random.default_rng(seed=random_seed)
        with tf.name_scope(v_scope):
            with self.strategy.scope():
                self.model = model or self.create_model(**kwargs)

    def create_model(self):
        """
        Create a model. Scopes automatically applied.
        """
        return None

    def select_strategy(self, strategy):
        """
        Select either the given strategy or the default strategy.
        """
        return strategy or tf.distribute.get_strategy()

    # Subclass must override
    def train_step(self) -> dict:
        """
        Single epoch of training.
        """
        return {}

    def random_set(self, x: np.ndarray, y: np.ndarray, epoch_factor=1) -> tuple[np.ndarray]:
        p1 = np.concatenate([self.rng.permutation(x.shape[0]) for _ in range(epoch_factor)])
        p2 = np.concatenate([self.rng.permutation(x.shape[0]) for _ in range(epoch_factor)])
        breakpoint()
        return x[p1], x[p2], y[p1], y[p2]

    @_run_tf_fn(print_time=True)
    def fit(self, *args, epochs=100, early_stop=True, min_delta=0, patience=3, **kwargs):
        """
        Train the model based on the given parameters. Extra arguments are passed to train_step.
        @param epochs: epochs to train for.
        @param min_delta: Minimum change required to qualify as an improvement.
        @param patience: How many epochs with no improvement before giving up. patience=0 disables.
        """
        patience = patience or epochs + 1  # If patience==0, do not early stop
        if patience < 1:
            raise ValueError('Patience value must be >1.')
        wait = 0
        best_weights = self.model.get_weights()
        for i in range(epochs):
            start = time.time()
            if not self.quiet:
                print(f'Epoch {i + 1}:')
            this_history = self.train_step(*args, **kwargs)
            if not self.quiet:
                print(f'Epoch time: {time.time() - start}')
            self.history = {k: v + this_history[k] for k, v in self.history.items()} if \
                self.history else this_history
            if not early_stop or i == 0:
                continue
            prev_best = min(self.history['loss'][:-1])
            this_loss = self.history['loss'][-1]
            if math.isnan(this_loss):  # NaN detection
                print('Stopping due to numerical instability, divergence to NaN')
                self.model.set_weights(best_weights)
                break
            if this_loss < prev_best - min_delta:  # Early stopping
                best_weights = self.model.get_weights()
                wait = 0
            else:
                wait += 1
            if wait >= patience:
                print('Stopping early due to lack of improvement!')
                self.model.set_weights(best_weights)
                break
        return self.history

    def transform(self, data: np.ndarray):
        """
        Transform the given data.
        """
        return data

    def save(self, path: str, model=None):
        """
        Save the model to the given path.
        @param path: path to save to.
        """
        try:
            os.makedirs(path)
        except FileExistsError:
            print("WARN: Directory exists, overwriting...")
            shutil.rmtree(path)
            os.makedirs(path)
        model = model or self.model
        model.save(os.path.join(path, 'model.h5'))
        with open(os.path.join(path, 'distance.pkl'), 'wb') as f:
            pickle.dump(self.distance, f)
        with open(os.path.join(path, 'embed_dist.txt'), 'w') as f:
            f.write(self.embed_dist)
        if self.history:
            with open(os.path.join(path, 'history.json'), 'w') as f:
                json.dump(self.history, f)

    @classmethod
    def load(cls, path: str, v_scope: str, strategy=None, model=None, **kwargs):
        """
        Load the model from the filesystem.
        """
        contents = os.listdir(path)
        if not model:
            if 'model.h5' not in contents:
                raise ValueError('Model save file is necessary for loading a ComparativeModel!')
            strategy = strategy or tf.distribute.get_strategy()
            with tf.name_scope(v_scope):
                with strategy.scope():
                    model = tf.keras.models.load_model(os.path.join(path, 'model.h5'))

        if 'distance.pkl' not in contents:
            print('Warning: distance save file missing!')
            dist = None
        else:
            with open(os.path.join(path, 'distance.pkl'), 'rb') as f:
                dist = pickle.load(f)
        if 'embed_dist.txt' not in contents:
            print('Warning: embedding distance save file missing, assuming Euclidean')
            dist = 'euclidean'
        else:
            with open(os.path.join(path, 'embed_dist.txt'), 'r') as f:
                embed_dist = f.read().strip()
        history = None
        if 'history.json' in contents:
            with open(os.path.join(path, 'history.json'), 'r') as f:
                history = json.load(f)
        return cls(v_scope=v_scope, dist=dist, model=model, strategy=strategy, history=history,
                   embed_dist=embed_dist, **kwargs)


class ComparativeEncoder(ComparativeModel):
    """
    Generic comparative encoder that can fit to data and transform sequences.
    """
    def __init__(self, model: tf.keras.Model, v_scope='encoder', **kwargs):
        """
        @param encoder: TensorFlow model that must support .train() and .predict() at minimum.
        @param dist: distance metric to use when comparing two sequences.
        """
        properties = {
            'input_shape': model.layers[0].output.shape[1:],
            'input_dtype': model.layers[0].dtype,
            'repr_size': model.layers[-1].output.shape[1],
            'depth': len(model.layers),
        }
        self.encoder = model
        super().__init__(v_scope, properties=properties, **kwargs)

    def create_model(self, loss='mse'):
        inputa = tf.keras.layers.Input(self.properties['input_shape'], name='input_a',
                                       dtype=self.properties['input_dtype'])
        inputb = tf.keras.layers.Input(self.properties['input_shape'], name='input_b',
                                       dtype=self.properties['input_dtype'])
        # Set embedding distance calculation layer
        if self.embed_dist.lower() == 'euclidean':
            dist_cl = EuclideanDistanceLayer()
        elif self.embed_dist.lower() == 'hyperbolic':
            dist_cl = HyperbolicDistanceLayer(init_scaling=True)
        else:
            raise ValueError('Invalid embedding distance provided!')
        distances = dist_cl(
            self.encoder(inputa),
            self.encoder(inputb),
        )
        if loss == 'mse':
            loss_kwargs = {'loss': 'mse', 'metrics': ['mae']}
        elif loss == 'corr_coef':
            loss_kwargs = {'loss': self.correlation_coefficient_loss}
        else:
            loss_kwargs = {'loss': loss}
        comparative_model = tf.keras.Model(inputs=[inputa, inputb], outputs=distances)
        comparative_model.compile(optimizer='adam', **loss_kwargs)
        return comparative_model

    @classmethod
    def from_model_builder(cls, builder: ModelBuilder, repr_size=None, norm_type='clip', embed_dist='euclidean',
                           **kwargs):
        """
        Initialize a ComparativeEncoder from a ModelBuilder object. Easy way to propagate the
        distribute strategy and variable scope. Also automatically adds a clip_norm for hyperbolic.
        """
        encoder = builder.compile(repr_size=repr_size, norm_type=norm_type, embed_space=embed_dist)
        return cls(encoder, strategy=builder.strategy, v_scope=builder.v_scope, embed_dist=embed_dist, **kwargs)

    @staticmethod
    def correlation_coefficient_loss(y_true, y_pred):
        """
        Correlation coefficient loss function for ComparativeEncoder.
        """
        x, y = y_true, y_pred
        mx, my = M.reduce_mean(x), M.reduce_mean(y)
        xm, ym = x - mx, y - my
        r_num = M.reduce_sum(M.multiply(xm, ym))
        r_den = M.sqrt(M.multiply(M.reduce_sum(M.square(xm)), M.reduce_sum(M.square(ym))))
        r = r_num / r_den
        r = M.maximum(M.minimum(r, 1.0), -1.0)
        return 1 - r

    # pylint: disable=arguments-differ
    def train_step(self, batch_size: int, data: np.ndarray, distance_on: np.ndarray, epoch_factor=1):
        """
        Train a single randomized epoch on data and distance_on.
        @param data: data to train model on.
        @param distance_on: np.ndarray of data to use for distance computations. Allows for distance
        to be based on secondary properties of each sequence, or on a string representation of the
        sequence (e.g. for alignment comparison methods).
        @param batch_size: batch size for TensorFlow.
        @param jobs: number of CPU jobs to use.
        @param chunksize: chunksize for Python multiprocessing.
        """
        # It's common to input pandas series from Dataset instead of numpy array
        data = data.to_numpy() if isinstance(data, pd.Series) else data
        distance_on = distance_on.to_numpy() if isinstance(distance_on, pd.Series) else distance_on
        x1, x2, y1, y2 = self.random_set(data, distance_on, epoch_factor=epoch_factor)

        y = self.distance.transform_multi(y1, y2)

        train_data = _prepare_tf_dataset({'input_a': x1, 'input_b': x2}, y, batch_size)

        return self.model.fit(train_data, epochs=1).history

    def fit(self, *args, distance_on=None, **kwargs):
        distance_on = distance_on if distance_on is not None else args[1]
        super().fit(*args, distance_on, **kwargs)

    @_run_tf_fn()
    def transform(self, data: np.ndarray, batch_size: int) -> np.ndarray:
        """
        Transform the given data into representations using trained model.
        @param data: np.ndarray containing all sequences to transform.
        @param batch_size: Batch size for .predict().
        @return np.ndarray: Representations for all sequences in data.
        """
        dataset = _prepare_tf_dataset(data, None, batch_size)
        return self.encoder.predict(dataset, batch_size=batch_size)

    def save(self, path: str):
        super().save(path, model=self.encoder)

    @classmethod
    def load(cls, path: str, v_scope='encoder', **kwargs):
        custom_objects = {'correlation_coefficient_loss': cls.correlation_coefficient_loss}
        with tf.keras.utils.custom_object_scope(custom_objects):
            return super().load(path, v_scope, **kwargs)

    def summary(self):
        """
        Prints a summary of the encoder.
        """
        self.encoder.summary()


class Decoder(ComparativeModel):
    """
    Abstract Decoder for encoding distances.
    """
    def __init__(self, v_scope='decoder', embed_dist_args=None, **kwargs):
        super().__init__(v_scope, **kwargs)
        embed_dist_args = embed_dist_args or {}
        if self.embed_dist == 'hyperbolic':
            self.embed_dist_calc = Hyperbolic(**embed_dist_args)
        elif self.embed_dist == 'euclidean':
            self.embed_dist_calc = Euclidean(**embed_dist_args)
        elif self.embed_dist == 'cosine':
            self.embed_dist_calc = Cosine(**embed_dist_args)
        else:  # Should never happen
            raise ValueError(f'Invalid embedding distance for decoder: {self.embed_dist}.')

    def random_distance_set(self, encodings: np.ndarray, distance_on: np.ndarray, epoch_factor=1):
        """
        Create a random set of distance data from the inputs.
        """
        x1, x2, y1, y2 = self.random_set(encodings, distance_on, epoch_factor=epoch_factor)

        if not self.quiet:
            print(f'Calculating embedding distances')
        x = self.embed_dist_calc.transform_multi(x1, x2)
        if not self.quiet:
            print('Calculating true distances')
        y = self.distance.transform_multi(y1, y2)
        return x, y

    def evaluate(self, encodings: np.ndarray, distance_on: np.ndarray, sample_size=None):
        """
        Evaluate the performance of the model by seeing how well we can predict true sequence
        dissimilarity from encoding distances.
        @param sample_size: Number of sequences to use for evaluation. All in dataset by default.
        @return np.ndarray, np.ndarray: predicted distances, true distances
        """
        sample_size = sample_size or len(encodings)
        x, y = self.random_distance_set(encodings, distance_on,
                                        epoch_factor=int(sample_size / len(encodings))+1)
        if not self.quiet:
            print('Predicting true distances...')
        x = self.transform(x)
        y = self.distance.invert_postprocessing(y)

        mse = mean_squared_error(y, x)
        r2 = r2_score(y, x)
        mape = mean_absolute_percentage_error(y, x)
        if not self.quiet:
            print(f'Mean squared error of distances: {mse}')
            print(f'R-squared correlation coefficient: {r2}')
            print(f'Mean absolute percentage error: {mape}')
        return x, y

    def save(self, path: str):
        super().save(path)
        with open(os.path.join(path, 'embed_dist_calc.pkl'), 'wb') as f:
            pickle.dump(self.embed_dist_calc, f)

    @staticmethod
    def load(path: str, v_scope='decoder', **kwargs):
        # pylint: disable=arguments-differ
        contents = os.listdir(path)
        if 'model.h5.pkl' in contents:
            obj = LinearDecoder.load(path)
        else:
            obj = DenseDecoder.load(path, **kwargs)
        with open(os.path.join(path, 'embed_dist_calc.pkl'), 'rb') as f:
            obj.embed_dist_calc = pickle.load(f)
        return obj


class _LinearRegressionModel(LinearRegression):
    def save(self, path: str):
        with open(path + '.pkl', 'wb') as f:
            pickle.dump(self, f)


class LinearDecoder(Decoder):
    """
    Linear model of a decoder. Useful with correlation coefficient loss.
    """
    # pylint: disable=arguments-differ
    def create_model(self):
        return _LinearRegressionModel()

    def fit(self, encodings: np.ndarray, distance_on: np.ndarray, *args, **kwargs):
        """
        Fit the LinearDecoder to the given data.
        """
        # It's common to input pandas series from Dataset instead of numpy array
        distance_on = distance_on.to_numpy() if isinstance(distance_on, pd.Series) else distance_on
        x, y = self.random_distance_set(encodings, distance_on, *args, **kwargs)
        self.model.fit(x.reshape((-1, 1)), y)

    def transform(self, data: np.ndarray):
        """
        Transform the given data.
        """
        return self.distance.invert_postprocessing(self.model.predict(data.reshape(-1, 1)))

    @classmethod
    def load(cls, path: str, **kwargs):
        with open(os.path.join(path, 'model.h5.pkl'), 'rb') as f:
            model = pickle.load(f)
        return super(Decoder, cls).load(path, 'decoder', model=model, **kwargs)


class DenseDecoder(Decoder):
    """
    Decoder model to convert generated distances into true distances.
    """
    def __init__(self, batch_size: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.batch_size = batch_size

    # pylint: disable=arguments-differ
    def create_model(self):
        dec_input = tf.keras.layers.Input((1,))
        x = tf.keras.layers.Dense(10, activation='relu')(dec_input)
        x = tf.keras.layers.Dense(10, activation='relu')(x)
        x = tf.keras.layers.Dropout(rate=.1)(x)
        x = tf.keras.layers.Dense(10, activation='relu')(x)
        x = tf.keras.layers.Dense(1, activation='relu')(x)
        decoder = tf.keras.Model(inputs=dec_input, outputs=x)
        decoder.compile(optimizer='adam', loss=tf.keras.losses.MeanAbsolutePercentageError())
        return decoder

    def fit(self, *args, epochs=25, **kwargs):
        """
        The decoder is probably an afterthought, 25 epochs seems like a sensible default to avoid
        adding too much overhead.
        """
        return super().fit(*args, epochs=epochs, **kwargs)

    def train_step(self, encodings: np.ndarray, distance_on: np.ndarray, epoch_factor=1):
        # It's common to input pandas series from Dataset instead of numpy array
        distance_on = distance_on.to_numpy() if isinstance(distance_on, pd.Series) else distance_on

        x, y = self.random_distance_set(encodings, distance_on, epoch_factor=epoch_factor)
        train_data = _prepare_tf_dataset(x, y, self.batch_size)

        return self.model.fit(train_data, epochs=1).history

    @_run_tf_fn()
    def transform(self, data: np.ndarray) -> np.ndarray:
        """
        Transform the given distances between this model's encodings into predicted true distances.
        """
        dataset = _prepare_tf_dataset(data, None, self.batch_size)
        return self.distance.invert_postprocessing(self.model.predict(dataset))

    def summary(self):
        """
        Prints a summary of the decoder.
        """
        self.model.summary()

    @classmethod
    def load(cls, path: str, v_scope='decoder', **kwargs):
        return super(Decoder, cls()).load(path, v_scope, **kwargs)




"""
Classes to help load and preprocess a dataset from the filesystem.
"""
import typing
import re
import matplotlib.pyplot as plt

import numpy as np
import pandas as pd
from Bio import SeqIO
import tensorflow as tf
from tqdm import tqdm

from .kmers import KMerCounter, Nucleotide_AA
tqdm.pandas()


class _SequenceTrimmer:
    """
    Helper class for Python multiprocessing library, allows for sequence trimming and padding
    to be multiprocessed with arbitrary trim lengths.
    """
    def __init__(self, length: int):
        self.length = length

    def trim(self, seq: str) -> str:
        """
        Performs trimming.
        """
        seq = seq[:self.length]
        return seq + ('N' * (self.length - len(seq)))


class LabelSeries(pd.Series):
    """
    Extends pd.Series, contains additional methods to generate masks based on labels.
    """
    _metadata = ['labels']

    @property
    def _constructor(self):
        return LabelSeries

    def label_mask(self, label: typing.Union[str, int], value: str) -> pd.Series:
        """
        Returns a mask for all values of this series where the given label equals the value.
        @param label: label to target search
        @param value: search query
        @return pd.Series: boolean mask
        """
        label = self.labels.index(label) if isinstance(label, str) else label
        if label == -1:
            raise ValueError('Label not in dataset!')
        return self.apply(lambda i: i[label] == value)

    def correct_length_mask(self) -> pd.Series:
        """
        Return a mask for all values of this series with a correct number of elements.
        @return pd.Series: boolean mask
        """
        return self.apply(len) == len(self.labels)

    def index_of_label(self, label: str) -> int:
        """
        Given a label, return the index of that label.
        @param label: label to search for
        @return int: index
        """
        return self.labels.index(label)


class DatasetBuilder:
    """
    Constructs a Dataset from input files.
    """
    def __init__(self, header_parser=None):
        """
        @param header_parser: HeaderParser object for header parsing
        """
        self.header_parser = header_parser

    @staticmethod
    def _read_fasta(path: str):
        """
        Reads a fasta file using BioPython and returns a list of tuples: (name, sequence).
        """
        headers, seqs = [], []
        with open(path, 'r') as f:
            for header, seq in SeqIO.FastaIO.SimpleFastaParser(f):
                headers.append(header)
                seqs.append(seq)
        return np.array(headers), np.array(seqs)

    @staticmethod
    def _dataset_decorator(cls_type):
        """
        Adds necessary _constructor and _constructor_sliced properties to new Dataset objects.
        """
        class DatasetDecorated(Dataset):
            """
            Defines _constructor and _constructor_sliced for Dataset.
            """
            @property
            def _constructor(self):
                return DatasetDecorated

            @property
            def _constructor_sliced(self):
                return cls_type
        return DatasetDecorated

    def from_fasta(self, paths: list[str], max_rows=None):
        """
        Factory function that builds a dataset from a fasta file. Reads in all sequences from all
        fasta files in list.
        @param paths: list of fasta file paths.
        @return dataset: new dataset object
        """
        raw_headers, seqs = [], []
        for i in paths:
            headers, s = self._read_fasta(i)
            raw_headers.append(headers)
            seqs.append(s)
        raw_headers = np.concatenate(raw_headers)[:max_rows or -1]
        seqs = np.concatenate(seqs)[:max_rows or -1]
        labels = self.header_parser(raw_headers) if self.header_parser else None
        cls = self._dataset_decorator(type(labels)) if labels is not None else Dataset
        return cls({'orig_seqs': seqs, 'seqs': seqs, 'raw_headers': raw_headers, 'labels': labels})


class Dataset(pd.DataFrame):
    """
    Useful class for handling sequence data. Underlying storage container is a pandas DataFrame.
    Columns:
    orig_seqs: raw, unprocessed sequence data. acts like a "backup" when performing transformations
    on sequences.
    seqs: sequences that can be transformed by built-in functions
    raw_headers: raw, unprocessed header data.
    labels: label data, present only if HeaderParser passed to DatasetBuilder or if manually added
    """
    @property
    def _constructor(self):
        return Dataset

    def add_labels(self, lbl_rows: pd.Series, lbl_cols: list[str]):
        """
        Add label data to the dataframe after dataset creation. Allows for other methods of label
        parsing. Returns a new Dataset with label data added.
        @param lbl_rows: _LabelSeries column.
        @param lbl_cols: list of labels represented by each index.
        @return Dataset: subclass of Dataset with label data added.
        """
        # pylint: disable=protected-access
        lbl_series_type = HeaderParser._lbl_series_decorator(lbl_cols)
        series = lbl_series_type(lbl_rows)
        dataset_type = DatasetBuilder._dataset_decorator(lbl_series_type)
        return dataset_type({'orig_seqs': self['orig_seqs'], 'seqs': self['seqs'],
                             'raw_headers': self['raw_headers'], 'labels': series})

    def drop_bad_headers(self):
        """
        Drop all elements which have a header of the wrong length. Returns new DataFrame.
        """
        assert self['labels'] is not None
        return self.loc[self['labels'].correct_length_mask()].dropna()

    def length_dist(self, progress=True):
        """
        Plots a histogram of the sequence lengths in this dataset.
        Helpful for selecting a trim length.
        @param progress: show progress bar during length calculations
        """
        plt.hist(self['seqs'].progress_apply(len) if progress else self['seqs'].apply(len))
        plt.show()

    def trim_seqs(self, length: int):
        """
        Trim all sequences to the given length and pad sequences with N which are too short.
        @param length: length to trim to.
        """
        trimmer = _SequenceTrimmer(length)
        self['seqs'] = self['orig_seqs'].apply(trimmer.trim)

    def replace_unknown_nucls(self):
        """
        Replace all unknown nucleotide base pairs in all sequences with N.
        """
        self['seqs'] = self['seqs'].apply(lambda i: re.sub('[^ATGCUN]', 'N', i))

    def gen_kmer_seqs(self, K=-1, jobs=1, chunksize=1, output_len=None, progress=True,
                      avoid_pad=True, avoid_oov=False, counter=None) -> list[np.ndarray]:
        """
        Convert all sequences to ordinal encoded kmer sequences.
        @param output_len: If given, trims and pads sequences to this length.
        @param avoid_pad: Adds 1 to all kmer values to dodge pad token. True by default.
        @param avoid_oov: Additionally adds 1 to all kmer values to dodge oov token. False by
        default.
        """
        if counter is None and K == -1:
            raise ValueError('Either K must be specified or a KMerCounter object must be passed!')
        counter = counter or KMerCounter(K, jobs=jobs, chunksize=chunksize, quiet=not progress)
        kmers = counter.kmer_sequences(self['seqs'].to_numpy())
        if avoid_pad or avoid_oov:  # Ensures no kmer values conflict with special tokens.
            kmers = [i + int(avoid_pad) + int(avoid_oov) for i in kmers]
        if not output_len:
            return kmers
        return tf.keras.utils.pad_sequences(kmers, maxlen=output_len, value=0)

    def aa_seqs(self, jobs=1, chunksize=1, progress=True, converter=None) -> list[str]:
        """
        Returns amino acid sequences based on the sliding window kmers method. Includes all possible
        interpretations of the underlying nucleotide sequences. Non-invertible.
        """
        converter = converter or Nucleotide_AA(jobs=jobs, chunksize=chunksize, progress=progress)
        return converter.transform(self['seqs'].to_numpy())

    def count_kmers(self, K=-1, jobs=1, chunksize=1, progress=True, counter=None) -> np.ndarray:
        """
        Count kmers for all sequences.
        @param K: Length of sequences to match.
        @param jobs: number of multiprocessing jobs
        @param chunksize: chunksize for multiprocessing
        @param progress: optional progress bar
        @return np.ndarray: counts of each kmer for all sequences
        """
        if counter is None and K == -1:
            raise ValueError('Either K must be specified or a KMerCounter object must be passed!')
        counter = counter or KMerCounter(K, jobs=jobs, chunksize=chunksize, quiet=not progress)
        return counter.kmer_counts(self['seqs'].to_numpy())


class HeaderParser:
    """
    HeaderParser class that can be extended to parse headers from FASTA files into labels.
    """
    def __init__(self, label_extractor: callable, label_cols: list[str]):
        """
        It's easily possible to create a new custom HeaderParser with a custom label_extractor
        function and a custom label_cols list. No subclassing necessary.
        @param label_extractor: function that takes in a raw header string and outputs a list of
        labels.
        @param label_cols: list of strings that store what each position in the label list
        represents.
        """
        self.label_extractor = label_extractor
        self.label_cols = label_cols

    @staticmethod
    def _lbl_series_decorator(label_cols: list[str]) -> type:
        """
        Hidden decorator function that returns a subclass of LabelSeries with the 'labels' attribute
        defined.
        @param label_cols: list of labels
        """
        class LabelSeriesDecorated(LabelSeries):
            """
            Defines _constructor for LabelSeries class.
            """
            labels = label_cols

            @property
            def _constructor(self):
                return LabelSeriesDecorated
        return LabelSeriesDecorated

    def __call__(self, data: list[str]) -> LabelSeries:
        """
        __call__ is used to actually build the header dataset.
        @param data: list of unparsed headers
        @return LabelSeries: custom pandas series for header data
        """
        cls = self._lbl_series_decorator(self.label_cols)
        return cls([self.label_extractor(i) for i in data])



def _silva_tax_extractor(header: str):
    return np.array(' '.join(header.split(' ')[1:]).split(';'))
SILVA_header_parser = HeaderParser(_silva_tax_extractor,
                                   ['Domain', 'Phylum', 'Class', 'Order', 'Family', 'Genus',
                                    'Species'])


def _covid_variant_extractor(header: str):
    return np.array([header.split('|')[2]], dtype=str)
COVID_header_parser = HeaderParser(_covid_variant_extractor, ['Variant'])




"""
Visualization of sequence representations.
"""
import typing
import matplotlib.pyplot as plt
import numpy as np
from .dataset_builder import Dataset


def repr_scatterplot(reprs: np.ndarray, title=None, alpha=.1, marker='.', figsize=(8, 6),
                     savepath=None, scale=None):
    """
    Create a simple scatterplot of sequence representations.
    Suggested alpha for SILVA dataset representations: .05
    @param reprs: array of sequence representations
    @param title: plot title
    @param alpha: alpha value for plot points
    @param marker: marker icon
    @param figsize: size of figure to generate
    @param savepath: path to save plot to.
    """
    x, y = np.array(reprs).T
    plt.figure(figsize=figsize)
    plt.scatter(x, y, alpha=alpha, marker=marker)
    if title:
        plt.title(title)
    if scale is not None:
        plt.xlim(-scale, scale)
        plt.ylim(-scale, scale)
    if savepath:
        plt.savefig(savepath)
    plt.show()


def reprs_by_ds_label(reprs: np.ndarray, ds: Dataset, label: typing.Union[str, int], *args,
                      **kwargs):
    """
    Scatterplot of representations colored by the values of a given label. Precondition: bad headers
    have been dropped. Needs args and kwargs for reprs_by_label.
    @param reprs: same as for reprs_by_label
    @param ds: dataset object with header data
    @param label: label to plot, passed as either an int index or string name
    @param *args: other positional arguments for reprs_by_label
    @param **kwargs: optional keyword arguments for reprs_by_label
    """
    tax = np.stack(ds['labels'])

    label_idx = label if isinstance(label, int) else ds['labels'].index_of_label(label)
    if label_idx == -1:
        raise ValueError('Given label not in dataset!')
    labels = tax[:, label_idx]
    reprs_by_label(reprs, labels, *args, **kwargs)


def reprs_by_label(reprs: np.ndarray, lbls: np.ndarray, title: str,
                   alpha=.1, marker='.', filter=None, savepath=None, mask=None,
                   scale=None, random_seed=None, **kwargs):
    # pylint: disable=redefined-builtin
    """
    Scatterplot of representations colored based on an array of categorical labels. Lower-level
    function.
    @param reprs: sequence representations
    @param lbls: np.ndarray object with label data
    @param title: plot title
    @param alpha: alpha value for plotted points
    @param marker: marker symbol
    @param filter: minimum number of sequences for a category to be plotted
    @param savepath: path to save to
    @param mask: boolean mask to apply to all arrays
    """
    if mask is not None:
        reprs = reprs[mask]
        lbls = lbls[mask]
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111)
    rng = np.random.default_rng(seed=random_seed)
    legend_labels = []
    for i in np.unique(lbls):  # For each unique value
        pop = reprs[lbls == i]  # All labels matching it
        if filter and len(pop) > filter:  # If we are filtering and have a sufficient count
            # Randomly sample down to the filter size (balances the plot)
            samp = rng.choice(len(pop), size=filter, replace=False)
            x, y = zip(*pop[samp])  # Prepare x, y for plotting
        elif not filter:  # If we aren't filtering, just prepare to plot
            x, y = zip(*pop)
        else:
            continue
        legend_labels.append(i)
        # Plot x, y with all given plot arguments
        ax.scatter(x, y, alpha=alpha, marker=marker, **kwargs)

    # Other plot stuff
    ax.set_title(title)
    leg = plt.legend(legend_labels, markerscale=1, borderpad=1)
    for lh in leg.legendHandles:  # Set alpha of legend so that we can see it
        lh.set_alpha(1)
    if scale is not None:
        plt.xlim(-scale, scale)
        plt.ylim(-scale, scale)
    if savepath:  # Save file if requested
        plt.savefig(savepath)
    plt.show()  # Show plot



"""
Output management functions. Private mmodule, should not be accessed outside of the package.
"""
import os
import contextlib


def suppress_output(func: callable):
    """
    Call the function with no arguments and suppress its output.
    """
    with open(os.devnull, 'w') as devnull:
        with contextlib.redirect_stderr(devnull):
            with contextlib.redirect_stdout(devnull):
                return func()




import os, logging
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "1"
import tensorflow as tf


# One of these should disable annoying warnings
logging.disable(logging.WARNING)
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
logging.getLogger('tensorflow').setLevel(logging.FATAL)
tf.get_logger().setLevel(logging.ERROR)
tf.autograph.set_verbosity(1)


DNA = ['A', 'C', 'G', 'T']
RNA = ['A', 'C', 'G', 'U']
AminoAcids = list("ACDEFGHIKLMNPQRSTVWY")



"""
Contains distance metrics used for training ComparativeEncoders.
"""
import multiprocessing as mp
import numpy as np
from tqdm import tqdm
from Bio.Align import PairwiseAligner
import Levenshtein
from rdkit.Chem import AllChem, DataStructs
from .kmers import KMerCounter


# pylint: disable=method-hidden
class Distance:
    """
    Abstract class representing a distance metric for two sequences.
    Downstream subclasses must implement transform.
    """
    def __init__(self, jobs=1, chunksize=1, quiet=False):
        self.jobs = jobs
        self.chunksize = chunksize
        self.quiet = quiet

    #pylint: disable=unused-argument
    def transform(self, pair: tuple) -> int:
        """
        Transform a pair of elements into a single integer distance between those elements. This can
        return a non-distance score if postprocessor() and invert_postprocessing are implemented.
        @param pair: two-element tuple containing elements to compute distance between.
        @return int: distance value
        """
        return 0

    def postprocessor(self, data: np.ndarray) -> np.ndarray:
        """
        Postprocess the output of transform() into a distance. Allows for metrics that don't
        explicitly calculate distance via postprocessing similarity scores/other results into
        distance. Must be invertible.
        @param data: np.ndarray
        @return np.ndarray
        """
        return data

    def transform_multi(self, y1: np.ndarray, y2: np.ndarray, silence=False) -> np.ndarray:
        """
        Transform two large arrays of data.
        """
        if self.jobs == 1:
            it = zip(y1, y2) if self.quiet or silence else tqdm(zip(y1, y2), total=len(y1))
            y = np.fromiter((self.transform(i) for i in it), dtype=np.float64)
        else:
            with mp.Pool(self.jobs) as p:
                it = p.imap(self.transform, zip(y1, y2), chunksize=self.chunksize)
                y = np.fromiter((it if self.quiet or silence else tqdm(it, total=len(y1))),
                                dtype=np.float64)
        y = self.postprocessor(y)  # Vectorized transformations are applied here
        return y

    def invert_postprocessing(self, data: np.ndarray) -> np.ndarray:
        """
        Inverts the postprocessing of data to return raw transform results.
        @param data: np.ndarray
        @return np.ndarray
        """
        return data


class VectorizedDistance(Distance):
    """
    Distance with a vectorized implementation. Takes advantage of multiprocessing and
    vectorization. transform() must take in pairs of single elements OR pairs of arrays of elements.
    """
    def transform_multi(self, y1, y2, silence=False):
        if self.jobs == 1:
            return self.postprocessor(self.transform((y1, y2)))
        y1_split = np.array_split(y1, len(y1) // self.chunksize)
        y2_split = np.array_split(y2, len(y1) // self.chunksize)
        with mp.Pool(self.jobs) as p:
            it = p.imap(self.transform, zip(y1_split, y2_split), chunksize=1)
            y = list(it if self.quiet or silence else tqdm(it, total=len(y1_split)))
            y = np.concatenate(y)
        y = self.postprocessor(y)
        return y


class Euclidean(VectorizedDistance):
    """
    Basic Euclidean distance implementation
    """
    def transform(self, pair: tuple) -> int:
        return np.linalg.norm(pair[0] - pair[1], axis=-1)


class Cosine(VectorizedDistance):
    """
    Cosine distance implementation.
    """
    def transform(self, pair: tuple) -> int:
        # Subtracting from 1 to convert similarity to distance
        return np.sum(pair[0] * pair[1], axis=-1) / (np.linalg.norm(pair[0], axis=-1) *
                                                     np.linalg.norm(pair[1], axis=-1))

    def postprocessor(self, data):
        return 1 - super().postprocessor(data)

    def invert_postprocessing(self, data):
        return 1 - super().invert_postprocessing(data)


class Hyperbolic(VectorizedDistance):
    """
    Computes hyperbolic distance between two arrays of points in Poincaré ball model.
    Numpy implementation.
    """
    def transform(self, pair):
        a, b = pair
        a = a.astype(np.float64)  # Both arrays must be the same type
        b = b.astype(np.float64)
        eps = np.finfo(np.float64).eps  # Machine epsilon
        sq_norm = lambda v: np.clip(np.sum(v ** 2), eps, 1 - eps)

        numerator = np.sum((a - b) ** 2, axis=-1)
        denominator_a = 1 - sq_norm(a)
        denominator_b = 1 - sq_norm(b)
        frac = numerator / (denominator_a * denominator_b)
        return np.arccosh(1 + 2 * frac)


class IncrementalDistance(VectorizedDistance):
    """
    Incrementally applies a regular K-Mers based distance metric over raw sequences.
    Use when not enough memory exists to fully encode a dataset into K-Mers with the specified K.
    Meant for use with VectorizedDistances. distance must have a transform() that can handle pairs
    of arrays of elements, and these arrays can have length 1.
    """
    def __init__(self, distance: VectorizedDistance, counter: KMerCounter):
        super().__init__()
        self.distance = distance
        self.counter = counter
        self.jobs = self.distance.jobs
        self.chunksize = self.distance.chunksize

    def transform(self, pair: tuple) -> int:
        pair = [i if isinstance(i, np.ndarray) else np.array([i]) for i in pair]
        kmer_pair = [self.counter.kmer_counts(i, jobs=1, chunksize=1, silence=True) for i in pair]
        return self.distance.transform(tuple(kmer_pair))


class EditDistance(Distance):
    """
    Normalized Levenshtein edit distance between textual DNA sequences.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aligner = Levenshtein.distance

    def transform(self, pair: tuple) -> int:
        return self.aligner(*pair) / max(map(len, pair))


class SmithWaterman(Distance):
    """
    Normalized alignment distance between two textual DNA sequences. Distance is computed from
    Smith-Waterman local alignment similarity scores. Deprecated, not recommended unless this
    legacy functionality is necessary.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aligner = PairwiseAligner()

    def transform(self, pair: tuple) -> int:
        return self.aligner.align(*pair).score / max(map(len, pair))

    def postprocessor(self, data: np.ndarray) -> np.ndarray:
        return 1 - super().postprocessor(data)

    def invert_postprocessing(self, data: np.ndarray) -> np.ndarray:
        return 1 - super().invert_postprocessing(data)


class CompoundDistance(Distance):
    """
    Distance between two chemical compounds.
    """
    def transform(self, pair: tuple):
        fp1 = AllChem.GetMorganFingerprintAsBitVect(pair[0], 2, nBits=1024)
        fp2 = AllChem.GetMorganFingerprintAsBitVect(pair[1], 2, nBits=1024)
        return DataStructs.TanimotoSimilarity(fp1, fp2)

    def postprocessor(self, data: np.ndarray) -> np.ndarray:
        return 1 - super().postprocessor(data)

    def invert_postprocessing(self, data):
        return 1 - super().invert_postprocessing(data)



"""
Module to help build encoders for ComparativeEncoder. It is recommended to use ModelBuilder.
"""
import tensorflow as tf
from .exceptions import IncompatibleDimensionsException


class AttentionBlock(tf.keras.layers.Layer):
    """
    Custom AttentionBlock layer that also contains normalization.
    Similar to the Transformer encoder block.
    """
    def __init__(self, embed_dim, num_heads, ff_dim):
        super().__init__()
        self.att = tf.keras.layers.MultiHeadAttention(num_heads=num_heads, key_dim=embed_dim)
        self.ffn = tf.keras.Sequential(
            [tf.keras.layers.Dense(ff_dim, activation="relu"), tf.keras.layers.Dense(embed_dim)]
        )
        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

    def call(self, inputs):
        """
        Calls attention, normalization, feed forward, and second normalization layers.
        """
        attn_output = self.att(inputs, inputs)
        out1 = self.layernorm1(inputs + attn_output)
        ffn_output = self.ffn(out1)
        return self.layernorm2(out1 + ffn_output)


def _one_hot_encoding(x, depth: int, **kwargs):
    return tf.one_hot(tf.cast(x, tf.int32), depth=depth, **kwargs)


class ModelBuilder:
    """
    Class that helps easily build encoders for a ComparativeEncoder model.
    """
    def __init__(self, input_shape: tuple, input_dtype=None, distribute_strategy=None,
                 v_scope='encoder'):
        """
        Create a new ModelBuilder object.
        @param input_shape: Shape of model input.
        @param input_dtype: Optional dtype for model input.
        @param distribute_strategy: strategy to use for distributed training. Defaults to training
        on a single GPU.
        """
        self.strategy = distribute_strategy or tf.distribute.get_strategy()
        self.v_scope = v_scope
        with tf.name_scope(v_scope):
            with self.strategy.scope():
                self.inputs = tf.keras.layers.Input(input_shape, dtype=input_dtype)
        self.current = self.inputs

    # pylint: disable=no-self-argument,not-callable
    def _apply_scopes(fn):
        def in_scopes(self, *args, **kwargs):
            __doc__ = fn.__doc__  # Set the documentation
            with tf.name_scope(self.v_scope):
                with self.strategy.scope():
                    return fn(self, *args, **kwargs)
        return in_scopes

    @classmethod
    def text_input(cls, vocab: list[str], embed_dim: int, max_len: int, v_scope='encoder',
                   **kwargs):
        """
        Factory function that returns a new ModelBuilder object which can receive text input. Adds a
        TextVectorization and an Embedding layer to preprocess string input data. Split happens
        along characters. Additional keyword arguments are passed to ModelBuilder constructor.

        @param vocab: Vocabulary to adapt TextVectorization layer to. String of characters with no
        duplicates.
        @param embed_dim: Size of embeddings to generate for each character in sequence. If None or
        not passed, defaults to one hot encoding of input sequences.
        @param max_len: Length to trim and pad input sequences to.
        @return ModelBuilder: Newly created object.
        """
        obj = cls((1,), input_dtype=tf.string, v_scope=v_scope, **kwargs)
        obj.text_vectorization(output_sequence_length=max_len,
                               output_mode='int',
                               vocabulary=vocab,
                               standardize=None,
                               split='character')
        obj.embedding(len(vocab) + 2, embed_dim)
        return obj

    @_apply_scopes
    def text_vectorization(self, *args, **kwargs):
        """
        Passes arguments directly to TextVectorization layer.
        """
        self.current = tf.keras.layers.TextVectorization(*args, **kwargs)(self.current)

    @_apply_scopes
    def one_hot_encoding(self, depth: int, **kwargs):
        """
        Add one hot encoding for the input. Input must be ordinally encoded data. Input will be
        casted to int32. Calls tf.one_hot().
        @param depth: number of categories to encode.
        """
        self.current = _one_hot_encoding(self.current, depth, **kwargs)

    @_apply_scopes
    def embedding(self, input_dim: int, output_dim: int, mask_zero=False, **kwargs):
        """
        Adds an Embedding layer to preprocess ordinally encoded input sequences.
        Arguments are passed directly to Embedding constructor.
        @param input_dim: Each input character must range from [0, input_dim).
        @param output_dim: Size of encoding for each character in the sequences.
        @param mask_zero: Whether to generate a mask for zero values in the input. Defaults to True.
        """
        self.current = tf.keras.layers.Embedding(input_dim, output_dim,
                                                 mask_zero=mask_zero,
                                                 **kwargs)(self.current)

    def summary(self):
        """
        Display a summary of the model as it currently stands.
        """
        tf.keras.Model(inputs=self.inputs, outputs=self.current).summary()

    def shape(self) -> tuple:
        """
        Returns the shape of the output layer as a tuple. Excludes the first dimension of batch size
        """
        return tuple(self.current.shape[1:])

    @_apply_scopes
    def clip_norm(self, clip_norm, **kwargs):
        self.current = tf.clip_by_norm(self.current, clip_norm=clip_norm, **kwargs)

    @_apply_scopes
    def div_by_magnitude(self, axis=-1, **kwargs):
        # min_scale = 1e-7
        # max_scale = 1 - 1e-3
        # radius = tf.Variable(1e-2, trainable=True, name='radius')
        # normalized_embeddings = tf.nn.l2_normalize(self.current, axis=axis, **kwargs)
        # self.current = normalized_embeddings * tf.clip_by_value(radius, min_scale, max_scale)
        magnitudes = tf.norm(self.current, axis=axis, keepdims=True, **kwargs)
        self.current = self.current / tf.maximum(magnitudes, tf.keras.backend.epsilon())

    @_apply_scopes
    def dynamic_norm_scaling(self, axis=-1, **kwargs):
        """
        Scale down the input such that the maximum absolute value is 1.
        """
        self.current = self.current / tf.reduce_max(tf.norm(self.current, axis=axis, **kwargs))

    @_apply_scopes
    def compile(self, repr_size=None, embed_space='euclidean', norm_type='clip') -> tf.keras.Model:
        """
        Create and return an encoder model.
        @param repr_size: Number of dimensions of output point (default 2 for visualization).
        @return tf.keras.Model
        """
        if repr_size:
            self.flatten()
            self.dense(repr_size, activation=None)  # Create special output layer
        if embed_space == 'hyperbolic':
            if norm_type == 'clip':
                self.clip_norm(1)
            elif norm_type == 'div_by_magnitude':
                self.div_by_magnitude()
            elif norm_type == 'scale_down':
                self.dynamic_norm_scaling()
            else:
                print('WARN: Empty/invalid norm_type, compiling hyperbolic model without normalization...')
        return tf.keras.Model(inputs=self.inputs, outputs=self.current)

    @_apply_scopes
    def custom_layer(self, layer: tf.keras.layers.Layer):
        """
        Add a custom layer to the model.
        @param layer: TensorFlow layer to add.
        """
        self.current = layer(self.current)

    @_apply_scopes
    def reshape(self, new_shape: tuple, **kwargs):
        """
        Add a reshape layer. Additional keyword arguments accepted.
        @param new_shape: tuple new shape.
        """
        self.current = tf.keras.layers.Reshape(new_shape, **kwargs)(self.current)

    def transpose(self, a=0, b=1, **kwargs):
        """
        Transposes the input with a Reshape layer over the two given axes (flips them).
        First dimension for batch size is not included.
        @param a: First axis to transpose, defaults to 0.
        @param b: Second axis to transpose, defaults to 1.
        """
        shape = list(self.shape())
        tmp = shape[b]
        shape[b] = shape[a]
        shape[a] = tmp
        self.reshape(tuple(shape), **kwargs)

    @_apply_scopes
    def flatten(self, **kwargs):
        """
        Add a flatten layer. Additional keyword arguments accepted.
        """
        self.current = tf.keras.layers.Flatten(**kwargs)(self.current)

    @_apply_scopes
    def dropout(self, rate, **kwargs):
        """
        Add a dropout layer. Additional keyword arguments accepted.
        @param rate: rate to drop out inputs.
        """
        self.current = tf.keras.layers.Dropout(rate=rate, **kwargs)(self.current)

    @_apply_scopes
    def dense(self, size: int, depth=1, activation='relu', **kwargs):
        """
        Procedurally add dense layers to the model.
        @param size: number of nodes per layer.
        @param depth: number of layers to add.
        @param activation: activation function to use (relu by default).
        Additional keyword arguments are passed to TensorFlow Dense layer constructor.
        """
        for _ in range(depth):
            self.current = tf.keras.layers.Dense(size, activation=activation,
                                                 **kwargs)(self.current)
            self.current = tf.keras.layers.BatchNormalization()(self.current)

    @_apply_scopes
    def conv1D(self, filters: int, kernel_size: int, output_size: int, **kwargs):
        """
        Add a convolutional layer.
        Output passes through feed forward layer with size specified by output_dim.
        @param filters: number of convolution filters to use.
        @param kernel_size: size of convolution kernel. Must be less than the first dimension of
        prior layer's shape.
        @param output_size: output size of the layer.
        @param activation: activation function.
        Additional keyword arguments are passed to TensorFlow Conv1D layer constructor.
        """
        if len(self.shape()) != 2:
            raise IncompatibleDimensionsException()
        if kernel_size >= self.shape()[0]:
            raise IncompatibleDimensionsException()

        self.current = tf.keras.layers.Conv1D(filters, kernel_size, activation='relu',
                                              **kwargs)(self.current)
        self.current = tf.keras.layers.MaxPooling1D()(self.current)
        # Removes extra dimension from shape
        self.current = tf.keras.layers.Flatten()(self.current)
        self.current = tf.keras.layers.BatchNormalization()(self.current)
        self.dense(output_size, activation='relu')

    @_apply_scopes
    def attention(self, num_heads: int, output_size: int):
        """
        Add an attention layer. Embeddings must be generated beforehand.
        @param num_heads: Number of attention heads.
        @param output_size: Output size.
        @param rate: Dropout rate for AttentionBlock.
        """
        if len(self.shape()) != 2:
            raise IncompatibleDimensionsException()

        self.current = AttentionBlock(self.shape()[1], num_heads, output_size)(self.current)
        self.current = tf.keras.layers.BatchNormalization()(self.current)

